# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os, sys, distutils.dir_util
from xplt import __version__


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Xplt"
copyright = "2022, Philipp Niedermayer (github.com/eltos)"
author = "Philipp Niedermayer (github.com/eltos)"
github_username = "eltos"
github_repository = "xplt"
version = __version__
release = version

# Project sources
root = os.path.abspath("..")
sys.path.insert(0, root)

# Auto API
autoapi_type = "python"
autoapi_dirs = ["../xplt"]
autoapi_ignore = ["*/.ipynb_checkpoints/*"]
autoapi_options = [
    "members",
    "inherited-members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
]
autoapi_python_class_content = "both"
autoapi_member_order = "groupwise"
autodoc_typehints = "description"
autoapi_add_toctree_entry = False


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "autoapi.extension",
    "sphinx.ext.githubpages",
    "myst_nb",
]

myst_enable_extensions = [
    "colon_fence",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Example notebooks
def np_example_notebooks_init(app, *args):
    global np_example_notebooks
    np_example_notebooks = distutils.dir_util.copy_tree(
        os.path.join(root, "examples"), os.path.join(app.srcdir, "examples")
    )


def np_example_notebooks_clean(*args):
    for file in np_example_notebooks:
        os.remove(file)


nb_execution_mode = "off"


def setup(app):
    app.connect("config-inited", np_example_notebooks_init)
    app.connect("build-finished", np_example_notebooks_clean)


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_show_sourcelink = False
html_static_path = ["_static"]
html_theme_options = {
    "show_nav_level": 2,
    "github_url": "https://github.com/eltos/xplt",
    "icon_links": [
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/xplt",
            "icon": "fa-solid fa-cube",
        },
    ],
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Methods for plotting lines

"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-11-08"


import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import os

from .base import XsuitePlot


def iter_elements(line):
    """Iterate over elements in line

    Yields (name, element, s_from, s_to) with

    """
    el_s0 = line.get_s_elements("upstream")
    el_s1 = line.get_s_elements("downstream")
    for name, el, s0, s1 in zip(line.element_names, line.elements, el_s0, el_s1):
        if s0 == s1:  # thin lense located at element center
            if hasattr(el, "length"):
                s0, s1 = (s0 + s1 - el.length) / 2, (s0 + s1 + el.length) / 2
        yield name, el, s0, s1


class KnlPlot(XsuitePlot):
    def __init__(
        self,
        line=None,
        *,
        knl=None,
        ax=None,
        filled=True,
        display_units=None,
        resolution=1000,
        line_length=None,
        **subplots_kwargs,
    ):
        """
        A plot for knl values along line

        :param line: Line of elements.
        :param knl: List of orders n to plot knl values for. If None, automatically determine from line.
        :param ax: An axes to plot onto. If None, a new figure is created.
        :param filled: If True, make a filled plot instead of a line plot.
        :param display_units: Dictionary with units for parameters. Supports prefix notation, e.g. 'bet' for 'betx' and 'bety'.
        :param resolution: Number of points to use for plot.
        :param line_length: Length of line (only required if line is None).
        :param subplots_kwargs: Keyword arguments passed to matplotlib.pyplot.subplots command when a new figure is created.

        """
        super().__init__(display_units=display_units)

        if knl is None:
            if line is None:
                raise ValueError("Either line or knl parameter must not be None")
            knl = range(
                max([e.order for e in line.elements if hasattr(e, "order")]) + 1
            )
        self.knl = knl
        if line is None and line_length is None:
            raise ValueError("Either line or line_length parameter must not be None")
        self.S = np.linspace(0, line_length or line.get_length(), resolution)
        self.filled = filled

        # Create plot
        if ax is None:
            _, ax = plt.subplots(**subplots_kwargs)
        self.fig, self.ax = ax.figure, ax
        self.ax.set(xlabel=self.label_for("s"), ylabel="$k_nl$")
        self.ax.grid()

        # create plot elements
        self.artists = []
        for n in self.knl:
            if self.filled:
                artist = self.ax.fill_between(
                    self.S,
                    np.zeros_like(self.S),
                    alpha=0.5,
                    label=self.label_for(f"k{n}l"),
                    zorder=3,
                )
            else:
                (artist,) = self.ax.plot(
                    [], [], alpha=0.5, label=self.label_for(f"k{n}l")
                )
            self.artists.append(artist)
        self.ax.plot(self.S, np.zeros_like(self.S), "k-", lw=1)
        self.ax.legend(ncol=5)

        # set data
        if line:
            self.update(line, autoscale=True)

    def update(self, line, autoscale=False):
        """
        Update the line data this plot shows

        :param line: Line of elements.
        :param autoscale: Whether or not to perform autoscaling on all axes
        :return: changed artists
        """
        # compute knl as function of s
        KNL = np.zeros((len(self.knl), self.S.size))
        for name, el, s0, s1 in iter_elements(line):
            if hasattr(el, "knl"):
                for i, n in enumerate(self.knl):
                    if n <= el.order:
                        mask = (s0 <= self.S) & (self.S < s1)
                        KNL[i, mask] += el.knl[n]

        # plot
        s = self.factor_for("s")
        changed = []
        for n, art, knl in zip(self.knl, self.artists, KNL):
            f = self.factor_for(f"k{n}l")
            if self.filled:
                art.get_paths()[0].vertices[1 : 1 + knl.size, 1] = f * knl
            else:
                art.set_data((s * self.S, f * knl))
            changed.append(art)
        if autoscale:
            if self.filled:  # At present, relim does not support collection instances.
                self.ax.update_datalim(
                    mpl.transforms.Bbox.union(
                        [a.get_datalim(self.ax.transData) for a in self.artists]
                    )
                )
            else:
                self.ax.relim()
            self.ax.autoscale()

        return changed



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Base methods for plotting

"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-11-08"


import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pint
import re


VOID = object()


def get(obj, val, default=VOID):
    """Get value from object"""
    try:
        return getattr(obj, val)
    except:
        try:
            return obj[val]
        except:
            if default is not VOID:
                return default
    raise AttributeError(f"{obj} does not provide an attribute or index '{val}'")


def style(kwargs, **default_style):
    """Return kwargs or defaults"""
    return dict(default_style, **(kwargs or {}))


def data_unit(p):
    """Return data unit of parameter p as used by xsuite"""
    # https://github.com/xsuite/xsuite/raw/main/docs/physics_manual/physics_man.pdf
    if re.fullmatch(r"k(\d+)l", p):
        return 1 if p == "k0l" else "m^-" + p[1:-1]
    # fmt: off
    units = dict(        
        
        ## particles
        ###################
        s='m',            # Reference accumulated path length
        x='m',            # Horizontal position
        px="1",           # Px / (m/m0 * p0c) = beta_x gamma /(beta0 gamma0)
        y='m',            # Vertical position
        py="1",           # Py / (m/m0 * p0c)
        delta="1",        # (Pc m0/m - p0c) /p0c
        ptau="1",         # (Energy m0/m - Energy0) / p0c
        pzeta="1",        # ptau / beta0
        rvv="1",          # beta / beta0
        rpp="1",          # m/m0 P0c / Pc = 1/(1+delta)
        zeta='m',         # (s - beta0 c t )
        tau='m',          # (s / beta0 - ct)
        mass0='eV',       # Reference rest mass
        q0='e',           # Reference charge
        p0c='eV',         # Reference momentum
        energy0='eV',     # Reference energy
        gamma0="1",       # Reference relativistic gamma
        beta0="1",        # Reference relativistic beta
        
        ## twiss
        ###################
        betx='m',         # Horizontal twiss beta-function
        bety='m',         # Vertical twiss beta-function
        alfx="1",         # Horizontal twiss alpha-function
        alfy="1",         # Vertical twiss alpha-function
        gamx='1/m',       # Horizontal twiss gamma-function
        gamy='1/m',       # Vertical twiss gamma-function
        mux="1",          # Horizontal phase advance
        muy="1",          # Vertical phase advance
        #muzeta
        qx="1",           # Horizontal tune qx=mux[-1]
        qy="1",           # Vertical tune qy=mux[-1]
        #qs
        dx='m',           # Horizontal dispersion $D_{x,y}$ [m]
        dy='m',           # $D_{x,y}$ [m]
        #dzeta
        dpx="1",
        dpy="1",
        circumference='m',
        T_rev='s',
        slip_factor="1",                 # eta
        momentum_compaction_factor="1",  # alpha_c = eta+1/gamma0^2 = 1/gamma0_tr^2
        #betz0
        
    )   
    # fmt: on

    if p not in units:
        raise NotImplementedError(f"Data unit for parameter {p} not implemented")
    return units.get(p)


def factor_for(var, to_unit):
    """Return factor to convert parameter into unit"""
    if var in ("X", "Y"):
        xy = var.lower()[-1]
        quantity = pint.Quantity(f"({data_unit(xy)})/({data_unit('bet'+xy)})^(1/2)")
    elif var in ("Px", "Py"):
        xy = var[-1]
        quantity = pint.Quantity(f"({data_unit('p'+xy)})*({data_unit('bet'+xy)})^(1/2)")
    else:
        quantity = pint.Quantity(data_unit(var))
    return (quantity / pint.Quantity(to_unit)).to("").magnitude


class XsuitePlot:
    def __init__(
        self,
        display_units=None,
    ):
        """
        Base class for plotting

        :param display_units: Dictionary with units for parameters. Supports prefix notation, e.g. 'bet' for 'betx' and 'bety'.
        """

        self._display_units = dict(
            dict(
                x="mm",
                y="mm",
                p="mrad",
                X="mm^(1/2)",
                Y="mm^(1/2)",
                P="mm^(1/2)",
                k0l="rad",
            ),
            **(display_units or {}),
        )

    def factor_for(self, p):
        """Return factor to convert parameter into display unit"""
        return factor_for(p, self.display_unit_for(p))

    def display_unit_for(self, p):
        """Return display unit for parameter"""
        prefix = p[:-1] if len(p) > 1 and p[-1] in "xy" else p
        if p in self._display_units:
            return self._display_units[p]
        if prefix in self._display_units:
            return self._display_units[prefix]
        return data_unit(p)

    def label_for(self, *pp, unit=True):
        """
        Return label for list of parameters, joining where possible

        :param pp: Parameter names
        :param unit: Wheather to include unit
        """

        def texify(label):
            if m := re.fullmatch(r"k(\d+)l", label):
                return f"k_{m.group(1)}l"
            return {
                "alf": "\\alpha",
                "bet": "\\beta",
                "gam": "\\gamma",
                "mu": "\\mu",
                "d": "D",
            }.get(label, label)

        def split(p):
            if p[-1] in "xy":
                return p[:-1], p[-1]
            return p, ""

        # split pre- and suffix
        prefix, _ = split(pp[0])
        display_unit = self.display_unit_for(pp[0])
        suffix = []
        for p in pp:
            pre, suf = split(p)
            suffix.append(suf)
            if pre != prefix or self.display_unit_for(p) != display_unit:
                # no common prefix or different units, treat separately!
                return " ,  ".join([self.label_for(p) for p in pp])
        suffix = ",".join(suffix)

        # build label
        label = "$"
        if prefix:
            label += texify(prefix)
            if suffix:
                label += "_{" + suffix + "}"
        else:
            label += suffix
        label += "$"
        if unit and display_unit:
            display_unit = pint.Unit(display_unit)
            if display_unit != pint.Unit("1"):
                label += f" / ${display_unit:~l}$"
        return label


class FixedLimits:
    """Context manager for keeping axis limits fixed while plotting"""

    def __init__(self, axis):
        self.axis = axis

    def __enter__(self):
        self.limits = self.axis.get_xlim(), self.axis.get_ylim()

    def __exit__(self, *args):
        self.axis.set(xlim=self.limits[0], ylim=self.limits[1])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Methods for plotting phase space distributions

"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-09-06"


import matplotlib as mpl
from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
import numpy as np

from .base import XsuitePlot, get, style, FixedLimits
from .util import normalized_coordinates, denormalized_coordinates

pairwise = np.c_


class PhaseSpacePlot(XsuitePlot):
    def __init__(
        self,
        particles=None,
        kind=None,
        plot="auto",
        *,
        ax=None,
        mask=None,
        masks=None,
        display_units=None,
        twiss=None,
        color=None,
        cmap="Blues" or "magma_r",
        projections="auto",
        projections_kwargs=None,
        mean=False,
        mean_kwargs=None,
        std=False,
        std_kwargs=None,
        percentiles=None,
        percentile_kwargs=None,
        grid=None,
        titles="auto",
        **subplots_kwargs,
    ):
        """
        A plot for phase space distributions

        :param particles: A dictionary with particle information
        :param kind: Defines the properties to plot.
                     This can be a nested list or a separated string or a mixture of lists and strings where
                     the first list level (or separator ``,``) determines the subplots,
                     and the second list level (or separator ``-``) determines coordinate pairs.
                     In addition, abbreviations for x-y-parameter pairs are supported (e.g. 'x' for 'x-px').
                     For normalized coordinates, use uppercase letters (e.g. 'X' for 'X-Px').

                     Examples:
                      - ``'x'``: single subplot with x-px phase space
                      - ``[['x', 'px']]``: same as above
                      - ``'x,x-y'``: two suplots the first with x-px and the second with x-y phase space
                      - ``[['x', 'px'], ['x', 'y']]``: same as above

        :param plot: Defines the type of plot. Can be 'auto', 'scatter' or 'hist'. Default is 'auto' for which the plot type is chosen automatically based on the number of particles.
        :param ax: A list of axes to plot onto, length must match the number of subplots. If None, a new figure is created.
        :param mask: An index mask to select particles to plot. If None, all particles are plotted.
        :param masks: List of masks for each subplot.
        :param display_units: Dictionary with units for parameters.
        :param twiss: Object holding twiss parameters (alfx, alfy, betx and bety) for calculation of normalized coordinates.
        :param color: Color of the scatter plot. If None, the color is determined by the color cycle.
        :param cmap: Colormap to use for the hist plot.
        :param projections: Add histogrammed projections onto axis. Can be True, False, "x", "y", "auto" or a list of these for each subplot
        :param projections_kwargs: Additional kwargs for histogram projection (step plot)
        :param mean: Whether to indicate mean of distribution with a cross marker. Boolean or list of booleans for each subplot.
        :param mean_kwargs: Additional kwargs for mean cross
        :param std: Whether to indicate standard deviation of distribution with an ellipse. Boolean or list of booleans for each subplot.
        :param std_kwargs: Additional kwargs for std ellipses.
        :param percentiles: List of percentiles (in percent) to indicate in the distribution with ellipses. Can also be a list of lists for each subplot.
        :param percentile_kwargs: Additional kwargs for percentile ellipses.
        :param grid: Tuple (ncol, nrow) for subplot layout. If None, the layout is determined automatically.
        :param titles: List of titles for each subplot or 'auto' to automatically set titles based on plot kind.
        :param subplots_kwargs: Keyword arguments passed to matplotlib.pyplot.subplots command when a new figure is created.


        """
        super().__init__(display_units=display_units)

        # parse kind string by splitting at commas and dashes and replacing abbreviations
        abbreviations = dict(
            x=("x", "px"),
            y=("y", "py"),
            z=("zeta", "delta"),
            X=("X", "Px"),
            Y=("Y", "Py"),
        )
        if kind is None:
            kind = "x,y,z" if twiss is None else "X,Y,z"
        if isinstance(kind, str):
            kind = kind.split(",")
        kind = list(kind)
        for i in range(len(kind)):
            if isinstance(kind[i], str):
                kind[i] = kind[i].split("-")
                # replace abbreviations elements by corresponding tuple
                if len(kind[i]) == 1 and kind[i][0] in abbreviations:
                    kind[i] = abbreviations[kind[i][0]]
                if len(kind[i]) != 2:
                    raise ValueError(
                        "Kind must only contain exactly two coordinates per subplot, "
                        f"but got {kind[i]}"
                    )
        self.kind = kind
        n = len(self.kind)

        # sanitize parameters
        if not hasattr(mean, "__iter__"):
            mean = n * [mean]
        if not hasattr(std, "__iter__"):
            std = n * [std]
        if percentiles is None or not hasattr(percentiles[0], "__iter__"):
            percentiles = n * [percentiles]
        if isinstance(projections, str) or not hasattr(projections, "__iter__"):
            projections = n * [projections]

        if len(mean) != n:
            raise ValueError(f"mean must be a boolean or a list of length {n}")
        if len(std) != n:
            raise ValueError(f"std must be a boolean or a list of length {n}")
        if len(percentiles) != n:
            raise ValueError(f"percentiles list must be flat or of length {n}")
        if grid and (len(grid) != 2 or grid[0] * grid[1] < n):
            raise ValueError(f"grid must be a tuple (n, m) with n*m >= {n}")
        if plot not in ["auto", "scatter", "hist"]:
            raise ValueError("plot must be 'auto', 'scatter' or 'hist'")

        self.plot = plot
        self.percentiles = percentiles
        self.projections = projections
        self.twiss = twiss

        # Create plot axes
        if ax is None:
            if grid:
                ncol, nrow = grid
            else:
                nrow = int(np.sqrt(n))
                while n % nrow != 0:
                    nrow -= 1
                ncol = n // nrow

            kwargs = style(subplots_kwargs, figsize=(4 * ncol, 4 * nrow))
            _, ax = plt.subplots(nrow, ncol, **kwargs)
        if not hasattr(ax, "__iter__"):
            ax = [ax]
        self.ax = ax
        self.fig = self.axflat[0].figure
        if len(self.axflat) < n:
            raise ValueError(f"Need {n} axes but got only {len(self.axflat)}")

        # Create distribution plots
        self.artists_scatter = [None] * n
        self.artists_hexbin = [()] * n
        self.artists_mean = [None] * n
        self.artists_std = [None] * n
        self.artists_percentiles = [()] * n
        self.ax_twin = [{} for i in range(n)]
        self.artists_twin = [{} for i in range(n)]
        self.artists_hamiltonian = [{} for i in range(n)]

        for i, ((a, b), ax) in enumerate(zip(self.kind, self.axflat)):

            # 2D phase space distribution
            ##############################

            self.artists_scatter[i] = ax.scatter([], [], s=4, color=color)
            self._hxkw = dict(cmap=cmap, rasterized=True)

            # 2D mean indicator
            if mean[i]:
                kwargs = style(mean_kwargs, color="k", marker="+", ms=8, zorder=100)
                (self.artists_mean[i],) = ax.plot([], [], **kwargs)

            # 2D std ellipses
            if std[i]:
                kwargs = style(std_kwargs, color="k", lw=1, ls="-", zorder=100)
                self.artists_std[i] = Ellipse([0, 0], 0, 0, fill=False, **kwargs)
                ax.add_artist(self.artists_std[i])

            # 2D percentile ellipses
            if percentiles[i]:
                self.artists_percentiles[i] = []
                for j, _ in enumerate(self.percentiles[i]):
                    kwargs = style(
                        percentile_kwargs,
                        color="k",
                        lw=1,
                        ls=(0, [5, 5] + [1, 5] * j),
                        zorder=100,
                    )
                    artist = Ellipse([0, 0], 0, 0, fill=False, **kwargs)
                    ax.add_artist(artist)
                    self.artists_percentiles[i].append(artist)

            # Axes title and labels
            if titles == "auto":
                title = self.title_for(a, b)
            elif titles is None:
                title = None
            else:
                title = titles[i]
            ax.set(title=title, xlabel=self.label_for(a), ylabel=self.label_for(b))
            ax.grid(alpha=0.5)

            # 1D histogram projections
            ###########################

            if self.projections[i]:
                kwargs = style(projections_kwargs, color="k", alpha=0.3, lw=1)
                for xy, yx in zip("xy", "yx"):
                    if self.projections[i] != yx:
                        # Create twin xy axis and artists
                        twin = ax.twinx() if xy == "x" else ax.twiny()
                        twin.set(**{f"{yx}ticks": [], f"{yx}lim": (0, 0.2)})
                        self.ax_twin[i][xy] = twin
                        (hist,) = twin.step([], [], **kwargs)
                        self.artists_twin[i][xy] = hist

        # set data
        if particles is not None:
            self.update(particles, mask=mask, masks=masks, autoscale=True)

    def update(self, particles, *, mask=None, masks=None, autoscale=False):
        """
        Update the data this plot shows

        :param particles: A dictionary with particle information
        :param mask: An index mask to select particles to plot. If None, all particles are plotted.
        :param masks: List of masks for each subplot.
        :param autoscale: Whether or not to perform autoscaling on all axes.
        :return: changed artists.
        """
        if masks is None:
            masks = [mask] * len(self.kind)
        elif mask is not None:
            raise ValueError("Only one of mask and masks can be set.")
        if len(masks) != len(self.kind):
            raise ValueError(f"masks must be a list of length {len(self.kind)}")

        for i, ((a, b), ax) in enumerate(zip(self.kind, self.axflat)):
            ax.autoscale(autoscale)
            # coordinates
            x = self.factor_for(a) * self._masked(particles, a, masks[i])
            y = self.factor_for(b) * self._masked(particles, b, masks[i])

            # statistics
            XY = np.array((x, y))
            XY0 = np.mean(XY, axis=1)
            UV = XY - XY0[:, np.newaxis]  # centered coordinates
            evals, evecs = np.linalg.eig(np.cov(UV))  # eigenvalues and -vectors

            # 2D phase space distribution
            ##############################

            plot = self.plot
            if plot == "auto":
                plot = "scatter" if len(x) <= 1000 else "hist"
            # scatter plot
            self.artists_scatter[i].set_visible(plot == "scatter")
            if plot == "scatter":
                self.artists_scatter[i].set_offsets(pairwise[x, y])
            # hexbin plot
            # remove old hexbin and create a new one (no update method)
            for artist in self.artists_hexbin[i]:
                artist.remove()
            self.artists_hexbin[i] = []
            if plot == "hist":
                # twice to mitigate https://stackoverflow.com/q/17354095
                hexbin_bg = ax.hexbin(x, y, mincnt=1, **self._hxkw)
                hexbin_fg = ax.hexbin(x, y, mincnt=1, edgecolors="none", **self._hxkw)
                self.artists_hexbin[i] = [hexbin_bg, hexbin_fg]

            # 2D mean indicator
            if self.artists_mean[i]:
                self.artists_mean[i].set_data(XY0)

            # 2D std indicator
            if self.artists_std[i]:
                w, h = 2 * np.sqrt(evals)
                self.artists_std[i].set(
                    center=XY0,
                    width=w,
                    height=h,
                    angle=np.degrees(np.arctan2(*evecs[1])),
                )

            # 2D percentile indicator
            if self.artists_percentiles[i]:
                # normalize distribution using eigenvalues and -vectors
                NN = np.dot(evecs.T, UV) / np.sqrt(evals)[:, np.newaxis]
                for j, p in enumerate(self.percentiles[i]):
                    # percentile in normalized distribution
                    e = np.percentile(np.sum(NN**2, axis=0), p) ** 0.5
                    w, h = 2 * e * np.sqrt(evals)
                    self.artists_percentiles[i][j].set(
                        center=XY0,
                        width=w,
                        height=h,
                        angle=np.degrees(np.arctan2(*evecs[1])),
                    )

            # 1D histogram projections
            ###########################

            project = self.projections[i]
            if project == "auto":
                project = len(x) >= 200

            for xy, v in zip("xy", (x, y)):
                if xy in self.artists_twin[i]:
                    hist = self.artists_twin[i][xy]
                    hist.set_visible(project)
                    if project:
                        # 1D histogram
                        counts, edges = np.histogram(v, bins=101)
                        counts = counts / len(v)
                        steps = (
                            np.append(edges, edges[-1]),
                            np.concatenate(([0], counts, [0])),
                        )
                        hist.set_data(steps if xy == "x" else steps[::-1])

            if autoscale:
                # ax.relim()  # At present, relim does not support collection instances.
                if plot == "scatter":
                    artists = [self.artists_scatter[i]]
                elif plot == "hist":
                    artists = self.artists_hexbin[i]
                else:
                    artists = []
                ax.update_datalim(
                    mpl.transforms.Bbox.union(
                        [a.get_datalim(ax.transData) for a in artists]
                    )
                )
                ax.autoscale()

    @property
    def axflat(self):
        return np.array(self.ax).flatten()

    def _masked(self, particles, prop, mask=None):
        """Get masked particle property"""

        if prop in ("X", "Px", "Y", "Py"):
            # normalized coordinates
            xy = prop.lower()[-1]
            coords = [get(particles, p) for p in (xy, "p" + xy)]
            delta = get(particles, "delta")
            X, Px = normalized_coordinates(*coords, self.twiss, xy, delta=delta)
            v = X if prop.lower() == xy else Px

        else:
            v = get(particles, prop)

        if mask is not None:
            v = v[mask]

        return np.array(v).flatten()

    def title_for(self, a, b):
        """
        Plot title for a given pair (a,b) of properties
        """
        titles = {  # Order of a,b does not matter
            "x-px": "Horizontal phase space",
            "y-py": "Vertical phase space",
            "zeta-delta": "Longitudinal phase space",
            "X-Px": "Normalized horizontal phase space",
            "Y-Py": "Normalized vertical phase space",
            "x-y": "Transverse profile",
        }
        return titles.get(f"{a}-{b}", titles.get(f"{b}-{a}", f"{a}-{b} phase space"))

    def axline(self, kind, val, also_on_normalized=True, subplots="all", **kwargs):
        """Plot a vertical or horizontal line for a given coordinate

        Args:
            kind (str): Phase space coordinate
            val (float): Value of phase space coordinate
            also_on_normalized (bool, optional): If true, also plot line for related (de-)normalized phase space coordinate
            subplots (list of int): Subplots to plot line onto. Defaults to all with matching coordinates.
            kwargs: Arguments for axvline or axhline

        """

        kwargs = style(kwargs, c="k")

        for i, (ab, ax) in enumerate(zip(self.kind, self.axflat)):
            if subplots is not "all" and i not in subplots:
                continue

            for a, line in zip(ab, (ax.axvline, ax.axhline)):
                if a == kind:
                    # same axis found, draw line
                    line(val * self.factor_for(kind), **kwargs)

                elif self.twiss is not None and kind.lower() == a.lower():
                    # related (de-)normalized axis found, transform and draw line
                    xy = kind.lower()[-1]
                    if kind in "x,y":  # x,y -> X,Y
                        v = normalized_coordinates(val, 0, self.twiss, xy)[0]
                    elif kind in "X,Y":  # X,Y -> x,y
                        v = denormalized_coordinates(val, 0, self.twiss, xy)[0]
                    elif kind in "px,py":  # px,py -> Px,Py
                        v = normalized_coordinates(0, val, self.twiss, xy)[1]
                    elif kind in "Px,Py":  # Px,Py -> px,py
                        v = denormalized_coordinates(0, val, self.twiss, xy)[1]
                    else:
                        continue
                    line(v * self.factor_for(a), **kwargs)

    def plot_hamiltonian_kobayashi(
        self,
        subplot,
        S,
        mu,
        *,
        q=None,
        extend=1,
        autoscale=True,
        separatrix=True,
        separatrix_kwargs=None,
        equipotentials=True,
        equipotentials_kwargs=None,
    ):
        """Plot separatrix and equipotential lines of kobayashi hamiltonian

        Args:
            subplot (int): Index of subplot
            S (float): Virtual sextupole strength in m^(-1/2)
            mu (float): Virtual sextupole phase in rad/2pi
            q (float, optional): Tune. Defaults to tune from twiss.
            extend (float, optional): Extend of separatrix and equipotential lines. If > 1 they are drawn beyond the the stable region.
            autoscale (bool, optional): Whether to autoscale axis or not
            separatrix (bool, optional): Plot separatrix. Defaults to True.
            separatrix_kwargs (dict, optional): Keyword arguments for separatrix line plot.
            equipotentials (bool, optional): Plot equipotential lines. Defaults to True.
            equipotentials_kwargs (dict, optional): Keyword arguments for equipotential line contour plot.
        """

        # TODO: provide option to plot for different particle delta (dispersion shifts and chromaticity shrinks the triangle)

        ax = self.axflat[subplot]
        a, b = self.kind[subplot]
        config = {
            "x-px": ("x", False),
            "y-py": ("y", False),
            "X-Px": ("x", True),
            "Y-Py": ("y", True),
        }
        if f"{a}-{b}" in config:
            xy, normalized = config[f"{a}-{b}"]
            swap = False
        elif f"{b}-{a}" in config:
            xy, normalized = config[f"{b}-{a}"]
            swap = True
        else:
            raise TypeError(f"Invalid plot kind {a}-{b} for kobayashi hamiltonian")

        # twiss parameters at plot location
        if self.twiss is None:
            raise ValueError("No twiss parameters provided during plot creation")
        alfx, betx, mux, x, px = [
            get(self.twiss, pre + xy).squeeze() for pre in ["alf", "bet", "mu", "", "p"]
        ]
        if q is None:
            q = get(self.twiss, "q" + xy).squeeze()

        # distance d to nearby 3rd order resonance r
        r = round(3 * q) / 3
        d = q - r
        # size of stable triangle
        h = 4 * np.pi * d / S
        # phase advance from virtual sextupole to here
        dmu = 2 * np.pi * (mux - mu)
        rotation = np.array([[+np.cos(dmu), np.sin(dmu)], [-np.sin(dmu), np.cos(dmu)]])

        def transform(XY):
            XY = np.tensordot(rotation, XY, 1)
            # match plot settings
            if not normalized:
                XY = np.array(denormalized_coordinates(*XY, self.twiss, xy))
            if swap:
                XY = XY[::-1]
            XY[0] *= self.factor_for(a)
            XY[1] *= self.factor_for(b)
            return XY

        lim = ax.dataLim
        with FixedLimits(ax):

            # plot separatrix
            if separatrix:
                kwarg = style(separatrix_kwargs, color="r", ls="--", label="Separatrix")

                t = extend
                for X, Y in (
                    [(-2, -2), (-2 * t, 2 * t)],
                    [(1 - 3 * t, 1 + 3 * t), (1 + t, 1 - t)],
                    [(1 - 3 * t, 1 + 3 * t), (-1 - t, -1 + t)],
                ):
                    X = h * np.array(X) / 2
                    Y = h * 3**0.5 * np.array(Y) / 2
                    ax.plot(*transform((X, Y)), **kwarg)

            # plot equipotential lines
            if equipotentials:
                kwargs = style(
                    equipotentials_kwargs, colors="lightgray", linewidths=1, alpha=0.5
                )

                # X = h * np.linspace(-5, 5, 500)
                X = extend * h * np.linspace(-1, 2, 500)
                Y = extend * h * 3**0.5 * np.linspace(-1, 1, 500)
                X, Y = np.meshgrid(X, Y)
                H = (3 * h * (X**2 + Y**2) + 3 * X * Y**2 - X**3) / h**3 / 4
                levels = np.linspace(0, extend, int(10 * extend)) ** 2

                ax.contour(*transform((X, Y)), H, levels=levels, **kwargs)

                ax.grid(False)

        # autoscale to separatrix
        if autoscale:
            edges = np.array(
                [(2 * h, -h, -h, 2 * h), (0, -(3**0.5) * h, 3**0.5 * h, 0)]
            )
            x, y = transform(edges)
            ax.dataLim = lim.union(
                [mpl.transforms.Bbox([[min(x), min(y)], [max(x), max(y)]])]
            )
            ax.autoscale()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Color definitions

Colors from https://arxiv.org/abs/2107.02270

"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-09-06"


import matplotlib as mpl


petroff_colors = [
    "#3f90da",
    "#ffa90e",
    "#bd1f01",
    "#94a4a2",
    "#832db6",
    "#a96b59",
    "#e76300",
    "#b9ac70",
    "#717581",
    "#92dadd",
]

# Colormap with 10 distinct colors
cmap_petroff = mpl.colors.ListedColormap(petroff_colors, "petroff")
cmap_petroff.set_under("none")
cmap_petroff.set_over("none")

# Gradient colormap
cmap_petroff_gradient = mpl.colors.LinearSegmentedColormap.from_list(
    "petroff_gradient", [petroff_colors[i] for i in (9, 0, 4, 2, 6, 1)]
)
cmap_petroff_gradient.set_under(petroff_colors[3])
cmap_petroff_gradient.set_over(petroff_colors[7])

# Bipolar colormap
cmap_petroff_bipolar = mpl.colors.LinearSegmentedColormap.from_list(
    "petroff_bipolar", [petroff_colors[i] for i in (2, 6, 1, 3, 9, 0, 4)]
)
cmap_petroff_bipolar.set_under(petroff_colors[5])
cmap_petroff_bipolar.set_over(petroff_colors[8])

# Cyclic colormap
cmap_petroff_cyclic = mpl.colors.LinearSegmentedColormap.from_list(
    "petroff_cyclic", [petroff_colors[i] for i in (3, 9, 0, 4, 2, 6, 1, 3)]
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Utility methods for accelerator physics

"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-11-15"


import numpy as np

from .base import get


def normalized_coordinates(x, px, twiss, xy, delta=0):
    """Convert physical to normalized coordinates

    Args:
        x: Physical x-coordinate in m
        px: Physical px-coordinate in rad
        twiss: Object or dict with twiss information in m and rad
        xy (str): Plane. Either "x" or "y".
        delta (float): Momentum deviation to account for dispersive orbit.

    Returns:
        Tuple of normalized coordinates (X, Px) in (m^(1/2), m^(1/2))
    """
    # substract (dispersive) closed orbit
    x = x - get(twiss, xy) - delta * get(twiss, "d" + xy)
    px = px - get(twiss, "p" + xy) - delta * get(twiss, "dp" + xy)
    # apply Floquet transform
    alf, bet = get(twiss, "alf" + xy), get(twiss, "bet" + xy)
    X = x / bet**0.5
    Px = alf * x / bet**0.5 + px * bet**0.5
    return X, Px


def denormalized_coordinates(X, Px, twiss, xy, delta=0):
    """Convert normalized to physical coordinates

    Args:
        X: Normalized X-coordinate in m^(1/2)
        Px: Normalized Px-coordinate in m^(1/2)
        twiss: Object or dict with local twiss information in m and rad
        xy (str): Plane. Either "x" or "y".
        delta (float): Momentum deviation to account for dispersive orbit.

    Returns:
        Tuple of physical coordinates (x, px) in (m, rad)
    """
    # apply Floquet transform
    alf, bet = get(twiss, "alf" + xy), get(twiss, "bet" + xy)
    x = X * bet**0.5
    px = -alf * X / bet**0.5 + Px / bet**0.5
    # add (dispersive) closed orbit
    x = x + get(twiss, xy) + delta * get(twiss, "d" + xy)
    px = px + get(twiss, "p" + xy) + delta * get(twiss, "dp" + xy)
    return x, px


def virtual_sextupole(tracker, particle_ref=None):
    """Determine virtual sextupole strength from twiss data

    The normalized strenght is defined as S = -1/2 * betx^(3/2) * k2l

    The implementation considers only normal sextupole components.

    Args:
        tracker: Tracker object with line and twiss methods
        particle_ref: Reference particle. Defaults to reference particle of tracker.

    Returns:
        Tuple (S, mu) with normalized strength in m^(-1/2) and phase in rad/2pi
    """

    # find sextupoles
    sextupoles, k2l = [], []
    for name, el in tracker.line.element_dict.items():
        if hasattr(el, "knl") and el.order >= 2 and el.knl[2]:
            sextupoles.append(name)
            k2l.append(el.knl[2])

    # twiss at sextupoles
    tw = tracker.twiss(
        method="4d",
        particle_ref=particle_ref,
        at_elements=sextupoles,
    )
    betx, mux = tw.betx, tw.mux

    # determine virtual sextupole
    Sn = -1 / 2 * betx ** (3 / 2) * k2l
    Stotal = np.sum(Sn * np.exp(3j * mux * 2 * np.pi))
    S = np.abs(Stotal)
    mu = np.angle(Stotal) / 3 / 2 / np.pi
    return S, mu



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"


__version__ = "0.1.7"


from .colors import *
from .line import *
from .twiss import *
from .phasespace import *


from . import hooks

try:
    hooks.register_matplotlib_options()
except:
    pass

try:
    hooks.register_pint_options()
except:
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Matplotlib setup

"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-09-06"

from xplt.colors import *


def register_matplotlib_options():
    """Register default options for matplotlib"""
    import matplotlib as mpl

    for cmap in (
        cmap_petroff,
        cmap_petroff_gradient,
        cmap_petroff_bipolar,
        cmap_petroff_cyclic,
    ):
        mpl.cm.register_cmap(cmap=cmap)
        cmap_r = cmap.reversed()
        cmap_r.name = cmap.name + "_r"
        mpl.cm.register_cmap(cmap=cmap_r)

    mpl.rcParams.update(
        {
            "figure.constrained_layout.use": True,
            "legend.fontsize": "x-small",
            "legend.title_fontsize": "small",
            "grid.color": "#DDD",
            "axes.prop_cycle": mpl.cycler(color=petroff_colors),
            # 'image.cmap': cmap_petroff_gradient,
        }
    )


def register_pint_options():
    """Register default options for pint"""
    import pint

    @pint.register_unit_format("l")
    def format_latex(unit, registry, **options):
        """Slightly modified latex formatter"""
        preprocessed = {
            r"\mathrm{{{}}}".format(u.replace("_", r"\_")): p for u, p in unit.items()
        }
        formatted = pint.formatter(
            preprocessed.items(),
            as_ratio=False,
            single_denominator=True,
            product_fmt=r" \cdot ",
            division_fmt=r"\frac[{}][{}]",
            power_fmt="{}^[{}]",
            parentheses_fmt=r"\left({}\right)",
            **options,
        )
        return formatted.replace("[", "{").replace("]", "}").replace("^{0.5}", "^{1/2}")

    pint.formatting.format_default = "l"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Methods for plotting twiss

"""

__author__ = "Philipp Niedermayer"
__contact__ = "eltos@outlook.de"
__date__ = "2022-11-08"


import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import os

from .base import XsuitePlot
from .line import KnlPlot


class TwissPlot(XsuitePlot):
    def __init__(
        self,
        twiss=None,
        kind="bet-dx,x+y",
        *,
        ax=None,
        line=None,
        line_kwargs=dict(),
        display_units=None,
        **subplots_kwargs,
    ):
        """
        A plot for twiss parameters and closed orbit

        :param kind: Defines the properties to plot.
                     This can be a nested list or a separated string or a mixture of lists and strings where
                     the first list level (or separator ``,``) determines the subplots,
                     the second list level (or separator ``-``) determines any twinx-axes,
                     and the third list level (or separator ``+``) determines plots.
                     In addition, abbreviations for x-y-parameter pairs are supported (e.g. 'bet' for 'betx+bety').

                     Examples:
                      - ``'bet-dx'``: single subplot with 'betx' and 'bety' on the left and 'dx' on the right axis
                      - ``[[['betx', 'bety'], ['dx']]]``: same as above
                      - ``'betx+alf,mu'``: two suplots the first with 'betx', 'alfx' and 'alfy' and the second with 'mux' and 'muy'
                      - ``[[['betx', 'alfx', 'alfy']], [['mux', 'muy']]]``: same as above

        :param twiss: Dictionary with twiss information
        :param ax: A list of axes to plot onto, length must match the number of subplots and optional line plot. If None, a new figure is created.
                   If required, twinx-axes will be added automatically.
        :param line: Line of elements. If given, adds a line plot to the top.
        :param line_kwargs: Keyword arguments passed to line plot.
        :param display_units: Dictionary with units for parameters. Supports prefix notation, e.g. 'bet' for 'betx' and 'bety'.
        :param subplots_kwargs: Keyword arguments passed to matplotlib.pyplot.subplots command when a new figure is created.

        """
        super().__init__(
            display_units=dict(
                dict(s="m", x="mm", y="mm", p="mrad", bet="m", d="m"),
                **(display_units or {}),
            )
        )

        # parse kind string
        if type(kind) is str:
            kind = kind.split(",")
        kind = list(kind)
        for i in range(len(kind)):
            if type(kind[i]) is str:
                kind[i] = kind[i].split("-")
            for j in range(len(kind[i])):
                if type(kind[i][j]) is str:
                    kind[i][j] = kind[i][j].split("+")
                k = 0
                while k < len(kind[i][j]):
                    if kind[i][j][k] in ("", "p", "alf", "bet", "gam", "mu", "d", "dp"):
                        kind[i][j].insert(k + 1, kind[i][j][k] + "y")
                        kind[i][j][k] += "x"
                        k += 1
                    k += 1
        self.kind = kind

        # Create plot axes
        if ax is None:
            _, ax = plt.subplots(
                len(self.kind) + (1 if line else 0), sharex="col", **subplots_kwargs
            )
        if not hasattr(ax, "__iter__"):
            ax = [ax]
        self.fig, self.ax = ax[0].figure, ax
        # Create line plot
        if line:
            self.lineplot = KnlPlot(line, ax=self.ax[0], **line_kwargs)
            self.lineplot.ax.set(xlabel=None)
            self.ax = self.ax[1:]
        self.ax_twin = [[] for a in self.ax]  # twinx-axes are created below
        # Format plot axes
        for a in self.ax:
            a.grid()
        self.ax[-1].set(xlabel="s / " + self.display_unit_for("s"))

        # create plot elements
        self.artists = []
        for i, ppp in enumerate(self.kind):
            self.artists.append([])
            legend = [], []

            for j, pp in enumerate(ppp):
                # axes
                if j == 0:
                    a = self.ax[i]
                else:  # create twinx-axes if required
                    a = self.ax[i].twinx()
                    a._get_lines.prop_cycler = self.ax[i]._get_lines.prop_cycler
                    self.ax_twin[i].append(a)
                a.set(ylabel=self.label_for(*pp))

                # create artists for traces
                self.artists[i].append([])
                for k, p in enumerate(pp):
                    (artist,) = a.plot([], [], label=self.label_for(p, unit=False))
                    self.artists[i][j].append(artist)
                    legend[0].append(artist)
                    legend[1].append(artist.get_label())

            if len(legend[0]) > 1:
                a.legend(*legend)

        # set data
        if twiss:
            self.update(twiss, autoscale=True)

    def update(self, twiss, *, autoscale=False, line=None):
        """
        Update the twiss data this plot shows

        :param twiss: Dictionary with twiss information.
        :param line: Line of elements.
        :param autoscale: Whether or not to perform autoscaling on all axes.
        :return: changed artists.
        """
        s = self.factor_for("s")
        changed = []
        for i, ppp in enumerate(self.kind):
            for j, pp in enumerate(ppp):
                a = self.ax[i] if j == 0 else self.ax_twin[i][j - 1]
                for k, p in enumerate(pp):
                    f = self.factor_for(p)
                    self.artists[i][j][k].set_data((s * twiss["s"], f * twiss[p]))
                    changed.append(self.artists[i][j][k])
                if autoscale:
                    a.relim()
                    a.autoscale()

        if line:
            self.lineplot.update(line, autoscale=autoscale)

        return changed
